"""
Exceptions for edx-username-changer plugin
"""


# pylint: disable=super-init-not-called
class UpdateFailedException(Exception):
    """
    Exception to throw when there is an update failure in username
    """

    def __init__(self, url, new_username):
        self.url = url
        self.new_username = new_username

    def __str__(self):
        return (
            f"Username update failed for username: {self.new_username}, url: {self.url}"
        )



"""
Django admin pages for edx-username-changer plugin
"""

from django.conf import settings
from django.contrib import admin
from django.contrib.auth import get_user_model

# pylint: disable=import-error
from common.djangoapps.student.admin import (
    UserAdmin as BaseUserAdmin,
)


User = get_user_model()


class UserAdmin(BaseUserAdmin):
    """
    Admin interface for the User model.
    """

    def get_readonly_fields(self, request, obj=None):
        """
        It will remove 'username' from the read-only fields to make it editable through the admin panel
        """
        readonly_fields = super().get_readonly_fields(request, obj)
        if settings.FEATURES.get("ENABLE_EDX_USERNAME_CHANGER") and obj:
            # pylint: disable=consider-using-generator
            return tuple([name for name in readonly_fields if name != "username"])
        return readonly_fields


# We must first un-register the User model since it was registered by edX's core code.
try:
    # pylint: disable=undefined-variable
    admin.site.unregister(User)
except NotRegistered:
    pass

admin.site.register(User, UserAdmin)



"""
This file contains celery tasks related to edx_username_changer plugin.
"""
from celery import shared_task  # pylint: disable=import-error

from django.contrib.auth import get_user_model

from openedx.core.djangoapps.django_comment_common.comment_client.user import (  # pylint: disable=import-error
    User as CommentUser,
)

from edx_username_changer.utils import (
    get_enrolled_course_ids,
    get_authored_threads_and_comments,
    update_comment_username,
    update_comment_user_username,
    update_thread_username,
)


COMMENT_TYPE = "comment"
THREAD_TYPE = "thread"
User = get_user_model()


@shared_task()
def task_update_username_in_forum(username):
    """
    Changes username in Discussion-Forum service
    """
    user = User.objects.get(username=username)
    comment_user = CommentUser.from_django_user(user)
    update_comment_user_username(comment_user, user.username)
    enrolled_course_ids = get_enrolled_course_ids(user)
    authored_items = get_authored_threads_and_comments(
        comment_user, enrolled_course_ids
    )

    for authored_item in authored_items:
        item_id = authored_item["id"]
        item_type = str(authored_item.get("type"))
        if item_type == THREAD_TYPE:
            update_thread_username(item_id, user.username)
        elif item_type == COMMENT_TYPE:
            update_comment_username(item_id, user.username)



"""
Utility methods for edx-username-changer plugin
"""

from django.contrib.auth import get_user_model
from django.db import transaction

from social_django.models import UserSocialAuth  # pylint: disable=import-error

from common.djangoapps.student.models import (  # pylint: disable=import-error
    CourseEnrollment,
)
from openedx.core.djangoapps.django_comment_common.comment_client.utils import (  # pylint: disable=import-error
    perform_request as perform_forum_request,
)
from openedx.core.djangoapps.django_comment_common.comment_client.thread import (  # pylint: disable=import-error
    Thread,
)
from openedx.core.djangoapps.django_comment_common.comment_client.comment import (  # pylint: disable=import-error
    Comment,
)

from edx_username_changer.exceptions import UpdateFailedException


User = get_user_model()


def update_user_social_auth_uid(old_username, new_username):
    """
    Changes uid in django-social-auth for OAuth based user accounts
    iff uid is based on username otherwise it doesn't make any effect
    """
    with transaction.atomic():
        UserSocialAuth.objects.filter(uid=old_username).update(uid=new_username)


def get_enrolled_course_ids(user):
    """
    Returns course ids of all the active enrollments of the provided user
    """
    return [
        str(enrollment.course_id)
        for enrollment in CourseEnrollment.enrollments_for_user(user)
    ]


def get_involved_threads(course_id, user_id):
    """
    Returns an iterator of all the discussion-forum threads against provided user and course
    """
    page = 0
    involved_threads = []
    while len(involved_threads) > 0 or page == 0:
        involved_threads = [
            Thread.find(id=thread["id"]).retrieve(
                with_responses=True, recursive=True, mark_as_read=False
            )
            for thread in Thread.search(
                {"course_id": course_id, "user_id": user_id, "page": page}
            ).collection
        ]
        yield from involved_threads
        page += 1


def get_authored_threads_and_comments(comment_user, enrolled_course_ids):
    """
    Returns an iterator of all the discussion-forum threads and comments of provided user and course
    """

    for course_id in enrolled_course_ids:
        involved_threads = get_involved_threads(course_id, comment_user.id)
        for thread in involved_threads:
            if thread["user_id"] == comment_user.id:
                yield thread.to_dict()

            children_to_scan = (
                thread.get("children", [])
                + thread.get("endorsed_responses", [])
                + thread.get("non_endorsed_responses", [])
            )

            while children_to_scan:
                child = children_to_scan.pop(0)
                children_to_scan.extend(child["children"])
                if child["user_id"] == comment_user.id:
                    yield child


def update_comment_user_username(comment_user, new_username):
    """
    Updates username for discussion-forum comment-users via Forum APIs
    """
    user_detail_url = comment_user.url_with_id(params={"id": comment_user.id})
    response_data = perform_forum_request(
        "put",
        user_detail_url,
        data_or_params={"username": new_username},
    )
    if response_data["username"] != new_username:
        raise UpdateFailedException(url=user_detail_url, new_username=new_username)


def update_thread_username(thread_id, new_username):
    """
    Updates username for discussion-forum threads via Forum APIs
    """
    thread_detail_url = Thread.url_with_id(params={"id": thread_id})
    response_data = perform_forum_request(
        "put",
        thread_detail_url,
        data_or_params={"username": new_username},
    )
    if response_data["username"] != new_username:
        raise UpdateFailedException(url=thread_detail_url, new_username=new_username)


def update_comment_username(comment_id, new_username):
    """
    Updates username for discussion-forum comments via Forum APIs
    """
    comment_detail_url = Comment.url_for_comments(params={"parent_id": comment_id})
    response_data = perform_forum_request(
        "put",
        comment_detail_url,
        data_or_params={"username": new_username},
    )
    if response_data["username"] != new_username:
        raise UpdateFailedException(url=comment_detail_url, new_username=new_username)



"""
Signals and Signal Handlers for edx-username-changer plugin
"""

from django.conf import settings

from common.djangoapps.util.model_utils import (  # pylint: disable=import-error
    get_changed_fields_dict,
)

from edx_username_changer.utils import update_user_social_auth_uid
from edx_username_changer.tasks import task_update_username_in_forum


def user_pre_save_callback(sender, **kwargs):
    """
    Pre-save signal handler of User model to store changed fields to be used later
    """
    if settings.FEATURES.get("ENABLE_EDX_USERNAME_CHANGER"):
        user = kwargs["instance"]
        fields_to_update = get_changed_fields_dict(user, sender)
        if "username" in fields_to_update:
            fields_to_update.update({"new_username": user.username})
            user._updated_fields = fields_to_update  # pylint: disable=protected-access


# pylint: disable=protected-access, unused-argument
def user_post_save_callback(sender, **kwargs):
    """
    Post-save signal handler of User model to update username throughout the application
    """
    if settings.FEATURES.get("ENABLE_EDX_USERNAME_CHANGER"):
        user = kwargs["instance"]
        if (
            hasattr(user, "_updated_fields")
            and user._updated_fields
            and {"username", "new_username"}.issubset(user._updated_fields)
        ):
            task_update_username_in_forum.delay(user._updated_fields["new_username"])
            update_user_social_auth_uid(
                user._updated_fields["username"], user._updated_fields["new_username"]
            )
            # pylint: disable=literal-used-as-attribute
            delattr(user, "_updated_fields")



import os
import setuptools
from edx_username_changer import __version__

with open("README.md", encoding="utf-8") as fh:
    long_description = fh.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setuptools.setup(
    name="edx-username-changer",
    version=__version__,
    author="MIT Office of Digital Learning",
    description="An edX plugin to change username of edx accounts through admin panel",
    license="GNU AFFERO GENERAL PUBLIC LICENSE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
    ],
    url="https://github.com/mitodl/edx-username-changer.git",
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    entry_points={
        "lms.djangoapp": [
            "edx_username_changer = edx_username_changer.apps:EdxUsernameChangerConfig",
        ],
        "cms.djangoapp": [
            "edx_username_changer = edx_username_changer.apps:EdxUsernameChangerConfig",
        ],
    },
)
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"



"""
App configuration for edx-username-changer plugin
"""

from django.apps import AppConfig

# pylint: disable=import-error, unused-import
from edx_django_utils.plugins.constants import PluginSignals, PluginSettings
from openedx.core.djangoapps.plugins.constants import (
    ProjectType,
    SettingsType,
)


class EdxUsernameChangerConfig(AppConfig):
    name = "edx_username_changer"
    verbose_name = "Open edX Username Changer"

    plugin_app = {
        PluginSignals.CONFIG: {
            ProjectType.LMS: {
                PluginSignals.RECEIVERS: [
                    {
                        PluginSignals.RECEIVER_FUNC_NAME: "user_pre_save_callback",
                        PluginSignals.SIGNAL_PATH: "django.db.models.signals.pre_save",
                        PluginSignals.SENDER_PATH: "django.contrib.auth.models.User",
                    },
                    {
                        PluginSignals.RECEIVER_FUNC_NAME: "user_post_save_callback",
                        PluginSignals.SIGNAL_PATH: "django.db.models.signals.post_save",
                        PluginSignals.SENDER_PATH: "django.contrib.auth.models.User",
                    },
                ],
            },
        },
    }
